from collections import deque
from datetime import datetime
from string import whitespace
from .config import ConversionFailure, ConvertValue
from .database import InfluxDatabase


class Metric(object):

    def __init__(self, entity, measurement, value):
        """
        Constructor for a single metric.

        :param entity: Identifier for the config entry which generated this measurement.
        :param measurement: String describing what this metric is recording.
        :param value: Raw value being record. The value will be serialized accordingly
                      when enqueued.
        """
        self.entity = entity
        self.timestamp = datetime.utcnow()
        self.measurement = measurement
        self.value = value

    @staticmethod
    def Sanitize(measurement):
        """
        Sanitize a measurement name. Removes whitespace and other separators and replaces
        them with an '_" (underscore).

        :param measurement: Measurement name
        :return: Sanitized string
        """
        for c in [' ', '.']:
            measurement = measurement.replace(c, '_')
        return measurement

    @staticmethod
    def TimeStamp(now=None):
        """
        Serialize a datetime into a ISO8601 date-time stamp.

        :param now: Optional datetime value. If None will default to utcnow().
        :return: Serialized date-time string.
        """
        now = now or datetime.utcnow()
        return now.strftime('%Y-%m-%dT%H:%M:%SZ')


class MetricPipeline(object):

    DEFAULT_BATCH_SIZE = 10

    def __init__(self, config, batchSize=DEFAULT_BATCH_SIZE, logger=None):
        """
        Constructor for the measurement pipeline.

        :param config: Full config from the Executor context.
        :param batchSize: Optional batch size. How many messages should be grouped into a single
                          call to the database.
        :param logger: Optional logger instance. If no instance is provided log messages will
                       be skipped.
        """
        self.config = config
        self.logger = logger
        self.batchSize = int(batchSize)
        self.queue = deque()
        self.database = None
        self.shutdown = False

    def __call__(self, metrics):
        self.Enqueue(metrics)

    def Enqueue(self, metrics):
        """
        Append a set of metrics onto the metric queue for sending to the backend database.
        Note that metrics are attempted after every iteration however if the database is
        unresponsive or unavailable then the metrics will remain in the queue until the
        database accepts them.

        :param metrics: A single metric or list of metrics which should be appended to the
                        end of the queue.
        :return: None
        """
        if self.shutdown:
            return
        if not isinstance(metrics, (list, set)):
            metrics = [metrics]
        for metric in metrics:
            if not isinstance(metric, Metric):
                if self.logger:
                    self.logger.warning('Invalid metric sent to the queue')
                continue
            try:
                metric.value = ConvertValue(
                    metric.value,
                    hint=self.config.GetField(metric.measurement))
            except KeyError:
                if self.logger:
                    self.logger.warning("Invalid metric '{}'".format(metric.measurement))
                continue
            except ConversionFailure:
                if self.logger:
                    self.logger.warning("Unable to convert measurement '{}' value '{}'"
                        .format(metric.measurement, metric.value))
                continue
            self.queue.append(metric)

    def Flush(self):
        """
        Flush the queue to the database. If any messages fail to send the queue will be unwound
        and retried on the next iteration pass. The inner loop will automatically handle
        batching based on the configured batch size.

        :return: None
        """
        if self.shutdown:
            return

        count = 0
        start = datetime.utcnow()
        status = True
        while status and not self.IsEmpty():
            metrics = []
            while len(metrics) < self.batchSize:
                try:
                    metrics.append(self.queue.popleft())
                except IndexError:
                    break
            points = []
            for metric in metrics:
                try:
                    points.append({
                        'measurement': Metric.Sanitize(metric.measurement),
                        'tags': self.config.GetTags(metric.entity),
                        'time': Metric.TimeStamp(metric.timestamp),
                        'fields': {'value': metric.value}})
                except KeyError:
                    continue
            try:
                if not self.database:
                    dbtype, config = self.config.GetDatabase()
                    if dbtype == 'influxdb':
                        self.database = InfluxDatabase(config, logger=self.logger, precision='ms')
                    else:
                        raise RuntimeError("Unknown database type '{}'".format(dbtype))

                self.database.Write(points)
            except RuntimeError as e:
                # Push any RuntimeErrors up to the main loop to handle. These usually indicate
                # that we need to crash.
                status = False
                raise e
            except Exception as e:
                if self.logger:
                    self.logger.error('Unhandled error sending metrics: {}'.format(e))
                status = False
            finally:
                if not status:
                    while metrics:
                        self.queue.appendleft(metrics.pop(-1))
            if status:
                count += len(points)

        stop = datetime.utcnow()
        if status and self.logger and count > 0:
            self.logger.debug("Uploaded '{}' metrics in {:.3f}ms"
                .format(count, (stop - start).total_seconds() * 1000))

    def IsEmpty(self):
        """
        Predicate to check if the message queue is empty.

        :return:
        """
        return len(self.queue) == 0

    def Reload(self, config):
        """
        Reload the configuration. This will trigger a full flush of the message queue to try
        and send the remaining messages to the database before re-initializing the connection.

        In the event of a failure the config will not be overwritten and reloaded.
        In the event of a successful flush, the database connection will be reconnected to handle
        a possible change in the database config.

        :param config: New config object from the Executor context.
        :return: None
        """
        if self.logger:
            self.logger.warning('Reloading metrics pipeline')
        try:
            self.Flush()
            if self.database:
                self.database.Close()
            self.database = None
            self.config = config
        except Exception as e:
            if self.logger:
                self.logger.warning('Error during reload: {}'.format(e))

    def Shutdown(self, crash=False):
        """
        Shutdown the message pipeline. This will flush the remaining metrics to the database
        before disconnecting the connection and setting the shutdown flag. If the database
        is not available this call wil not block and data could be lost.

        :return: None
        """
        try:
            self.Flush()
            if self.database:
                self.database.Close()
            self.database = None
        except Exception as e:
            if self.logger and not crash:
                self.logger.warning('Error during metric pipeline shutdown: {}'.format(e))
        self.shutdown = True



import argparse
import logging
from logging.handlers import TimedRotatingFileHandler
import os
import signal
import sys
from .config import Config, ConfigError
from .daemon import Daemonize
from .exceptions import ExecutorError
from .metrics import MetricPipeline
from .result import Result
from .utils import Callbacks, CloseDescriptor, Select, SetNonBlocking


class Executor(object):

    def __init__(self, args, root, callbacks=None):
        """
        Constructor for the Executor. It takes the argument parser args and the root
        configuration element from the caller and builds the corresponding data structures.

        :param args: Argument parser resultset.
        :param root: Root element in the configuration file determined by the caller.
        :param callbacks: Callback helper for handling user-generated sub-commands.
        """
        self.args = args
        self.callbacks = getattr(callbacks, 'callbacks', None)
        self.command = getattr(callbacks, 'command', None)
        self.action = getattr(args, 'command', None)
        self.config = Config(args.config, root)

        if not self.action or self.action == self.command:
            self.interval = int(args.interval)
            self.pidFile = args.pidfile
            self.logfile = args.logfile
            self.logger = self.SetupLogging(self.logfile, args.loglevel)
            self.context = Daemonize(args.daemon, self.logger,
                group=args.group,
                user=args.user)
            self.pipeline = MetricPipeline(self.config, logger=self.logger)
            self.__shutdown = False
            self.__rd, self.__wr = None, None
            self.__reload = False

    @staticmethod
    def Configure(parser):
        """
        Configure an argument parser with the necessary options for running the Executor.

        :param parser: Argument parser instance (can also be a sub-command)
        :return: None
        """
        parser.add_argument('--daemon', '-d', action='store_true', default=False,
            help='Daemonize the application (only supported on Linux).')
        parser.add_argument('--interval', '-i', type=int, default=10,
            help='Interval at which the runnable callback should be executed')
        parser.add_argument('--logfile', '-f', required=False,
            help='Path to the log file which should be used.')
        parser.add_argument('--loglevel', '-l',
            default='INFO',
            choices=['DEBUG', 'INFO', 'ERROR', 'FATAL', 'WARNING'],
            help='Log level used for logging.')
        parser.add_argument('--stdout', '-o', action='store_true', default=False,
            help='Add a StreamHandler to STDOUT for the executor application.')
        parser.add_argument('--pidfile', '-p', required=False,
            help='Path to where the PID file should be written')
        parser.add_argument('--uid', '-u', dest='user', required=False,
            help='After daemonizing run the process as the following user id.')
        parser.add_argument('--gid', '-g', dest='group', required=False,
            help='After daemonizing run the process as the following group id.')
        parser.add_argument('config',
            help='Path to the config file')

    def Notify(self):
        """
        Notify a sleeping iterator loop that an event occurred which must be handled.
        This could be a signal or another error that the system needs to wakeup and react
        too.

        :return:
        """
        if self.__wr:
            try:
                os.write(self.__wr, b'.')
            except (IOError, OSError):
                pass

    def Reload(self):
        """
        Reload the config from disk and re-initialize the Metrics pipeline if necessary.
        This is a blocking operation which will force flush the Mertics internal queue
        and could potentially cause a long block if the underlying database is slow or
        unresponsive.

        :return: None
        """
        if self.__reload:
            self.config.Reload()
            self.pipeline.Reload(self.config)

        self.__reload = False

    def Run(self, callback):
        """
        Execute the service. This will run the daemon context and fork the application
        if daemoniing has been selected otherwise it will create all of the resources
        and begin the calls of the callback at the given interval.

        :param callback: Runnable function callback which should be executed every
                         interval. This function MUST be idempotent and return a
                         value from the Result enumerable type.
        :return:
        """
        if self.action is not None and self.action != self.command:
            if self.action in self.callbacks:
                return self.callbacks[self.action](self.config, self.args)
            print("Unknown command: {}".format(self.action))
            raise SystemExit

        if not self.Start(validate=True):
            raise SystemExit

        crash = False
        with self.context:
            try:
                self.__rd, self.__wr = os.pipe()
                SetNonBlocking(self.__rd)
                SetNonBlocking(self.__wr)
            except (IOError, OSError):
                self.logger.error('Failed to initialize notify socket')
                raise

            try:
                if not self.Start():
                    raise SystemExit
                failures = 0
                while not self.__shutdown:
                    self.Reload()
                    try:
                        result = callback(self.config.GetRoot(), self.logger, self.pipeline)
                        if result == Result.CANCEL:
                            self.logger.info('Callback initiated shutdown')
                            self.__shutdown = True
                        elif result == result.FAILURE:
                            failures += 1
                            if failures > 10:
                                self.logger.error('Callback returned too many failures. Initiating shutdown.')
                                self.__shutdown = True
                        failures = 0
                        if not self.__shutdown:
                            self.pipeline.Flush()
                            if Select(self.__rd, [], self.interval):
                                try:
                                    os.read(self.__rd, 1)
                                except (IOError, OSError):
                                    pass
                    except KeyboardInterrupt:
                        self.logger.warning('Shutdown initiated')
                        self.__shutdown = True
            except RuntimeError as e:
                if self.logger:
                    self.logger.fatal('Fatal error in main loop: {}'.format(e))
                crash = True
            except Exception as e:
                if self.logger:
                    self.logger.error('Unexpected exception in main loop: {}'.format(e))
            finally:
                if self.__rd:
                    CloseDescriptor(self.__rd)
                if self.__wr:
                    CloseDescriptor(self.__wr)
                if self.pipeline:
                    try:
                        self.pipeline.Shutdown(crash=crash)
                    except Exception as e:
                        self.logger.warning('Error during database shutdown: {}'.format(e))

    def SetupLogging(self, logfile, loglevel, stdout=False):
        """
        Configure the log handler for the executor using the given logfile and
        loglevel. If no logfile is specified then logging will default to a StreamHandler
        to STDOUT.

        :param logfile: Path to the given logfile.
        :param loglevel: String representing the LogLevel that should be chosen.
        :param stdout: Flag indicating that a StreamHandler logging to STDOUT should be
                       created.
        :return: Application logger
        """
        logger = logging.getLogger('monitor.lib.Executor')

        if isinstance(loglevel, int):
            level = loglevel
        else:
            level = logging.WARNING
            if loglevel == 'DEBUG':
                level = logging.DEBUG
            elif loglevel == 'INFO':
                level = logging.INFO
            elif loglevel == 'WARNING':
                level = logging.WARNING
            elif loglevel == 'ERROR':
                level = logging.ERROR
            elif loglevel == 'CRITICAL':
                level = logging.CRITICAL

        if stdout or not logfile:
            console = logging.StreamHandler(stream=sys.stdout)
            console.setLevel(level)
            logger.addHandler(console)
        if logfile:
            handler = TimedRotatingFileHandler(
                logfile,
                backupCount=3,
                when='midnight')
            handler.setLevel(level)
            logger.addHandler(handler)

        logger.setLevel(level)
        logger.propagate = False
        return logger

    def SignalHandler(self, sig, frame):
        """
        Signal handler for the application. The SignalHandler will be installed after
        the application has successfully started up. In the case of a daemonize that
        signals will be registered after to launch has completed.

        :param sig: Signal integer received from the system.
        :return: None
        """
        self.logger.info('Signal received: {}'.format(sig))
        if sig in [signal.SIGINT, signal.SIGTERM]:
            self.__shutdown = True
        elif sig == signal.SIGHUP:
            self.__reload = True
        self.Notify()

    def Start(self, validate=False):
        """
        Start the Executor: this includes creating any logfile, directories, etc. If
        the validate flag is specified, the Start will only create directories and
        attempt check the configuration and other values.

        By default if the application is set to daemonize then Start(...) will be
        called again after daemonizing.

        :param validate: Signal that the call to Start(...) should not actually start
                         but only check the configuration for starting.
        :return:
        """
        if not self.config.IsLoaded():
            try:
                self.config.Load()
            except ConfigError as e:
                self.logger.error('Failed to load config: {}'.format(e.message))
                return False

        if self.logfile:
            path = os.path.basename(self.logfile)
            if not os.path.exists(path):
                try:
                    os.makedirs(path)
                except (IOError, OSError):
                    self.logger.error('Failed to create the log folder')
                    return False
            try:
                with open(self.logfile, 'ab'):
                    pass
            except (IOError, OSError):
                self.logger.error('Failed to create the log file')
                return False

        if self.pidFile:
            path = os.path.basename(self.pidFile)
            if not os.path.exists(path):
                try:
                    os.makedirs(path)
                except (IOError, OSError):
                    self.logger.error('Failed to create PID file directory')
                    return False
            if not validate:
                try:
                    with open(self.pidFile, 'wb') as handle:
                        handle.write(str(os.getpid()).encode('ascii'))
                except (IOError, OSError):
                    self.logger.error('Failed to write pid file')
                    return False

        if not validate:
            self.logger.debug('Installing signal handlers')
            signal.signal(signal.SIGHUP, self.SignalHandler)
            signal.signal(signal.SIGINT, self.SignalHandler)
            signal.signal(signal.SIGTERM, self.SignalHandler)

        return True


def Execute(callback, root, command='run', commands=None):
    """
    Create an Executor instance that runs for the given callback. This function
    acts as an all in one to create the Executor, build the command parser, and
    parse the given args. From there everything will be passed to the Executor
    to be run.

    This function will end the program based on the result of the Run(...) call
    and does not return back to the caller.

    :param callback: Callable object which will be called from the main service
                     every given interval.
    :param root: Root node in the config that lists all other entries. This field
                 should be listed under globals.
    :param command: String which should be used for the sub-command if the caller
                    supplies a callback for more sub-commands. This should
                    represent the command which triggers the daemonized polling
                    process to start.
    :param commands: Callback which can register more one-off sub-commands.
    :return: None
    """
    parser = argparse.ArgumentParser(sys.argv[0])
    callbacks = Callbacks(command)

    if commands is not None:
        if not callable(commands):
            raise ExecutorError('sub-command callbacks must be callable')
        parsers = parser.add_subparsers(dest='command',
            help='Available sub-command options')
        Executor.Configure(parsers.add_parser(command,
            help='Execute the daemonized polling process'))
        callbacks(parsers, commands)
    else:
        Executor.Configure(parser)

    args = parser.parse_args()

    if not hasattr(args, 'config'):
        parser.print_help()
        sys.exit(1)

    executor = Executor(args, root, callbacks=callbacks)
    executor.Run(callback)



from configparser import ConfigParser
import os
from .exceptions import MessageError


class ConfigError(MessageError):
    message = 'Invalid configuration'


class InvalidConfigError(ConfigError):
    pass


class ConversionFailure(MessageError):
    message = 'Invalid type conversion requested'


def ConvertBoolean(value):
    """
    Attempt to convert value to a boolean. If the value is not a possible boolean
    string the result will return None instead of True or False.

    :param value: Value which should be checked as a boolean.
    :return: True or False if the value is recognized, or None if it is not a valid
            boolean string.
    """
    value = str(value).lower()
    if value in ['yes', '1', 'true']:
        return True
    if value in ['no', '0', 'false']:
        return False
    return None


def ConvertHashType(value):
    """
    Attempt to convert a space separated series of key=value pairs into a dictionary
    of pairs. If any value fails to split successfully an error will be raised.

    :param value: Space delimited string of key-value pairs
    :return: Dictionary of key-value pairs.
    """
    collection = dict()
    for option in value.split():
        try:
            k, v = option.split('=')
        except ValueError:
            raise ConversionFailure("Invalid option '{}' for key-value pair: {}"
                .format(option, value))
        collection[k] = v.strip()
    return collection


def ConvertValue(value, hint=None):
    """
    Attempt to convert value to a supported type. Hint can be used to influence
    how the value is coerced. In the event of a failure an error will be raised.

    Supported hints are:
    - Array
    - Hash
    - Integer
    - Float
    - Boolean
    - String

    :param value: Value to be coerced.
    :param hint: Optional hint to indicate desired type.
    :return: Value coerced to type.
    """
    if isinstance(value, str):
        value = value.strip()
    try:
        if hint is not None:
            if hint == Config.ARRAY_TYPE:
                return value.split()
            elif hint == Config.HASH_TYPE:
                return ConvertHashType(value)
            elif hint == Config.INT_TYPE:
                if isinstance(value, str) and '.' in value:
                    return int(float(value))
                return int(value)
            elif hint == Config.BOOL_TYPE:
                return ConvertBoolean(value)
            elif hint == Config.FLOAT_TYPE:
                return float(value)
            elif hint == Config.STRING_TYPE:
                return value
            raise ConversionFailure
        else:
            b = ConvertBoolean(value)
            if b is not None:
                return b
            v = float(value)
            if isinstance(value, str) and '.' in value:
                return v
            return int(v)
    except (TypeError, ValueError):
        if hint is not None:
            raise ConversionFailure('Failed to convert the given value to the requested type')
        else:
            if ConvertBoolean(value) is None:
                return value
            return ConvertBoolean(value)


class Config(object):
    ARRAY_TYPE = 'array'
    BOOL_TYPE = 'bool'
    FLOAT_TYPE = 'float'
    HASH_TYPE = 'hash'
    INT_TYPE = 'int'
    STRING_TYPE = 'string'

    GLOBAL_SECTION = 'global'
    INFLUXDB_SECTION = 'influxdb'
    SUPPORTED_DATABASES = [INFLUXDB_SECTION]
    DATABASE_FIELDS = {
        INFLUXDB_SECTION: [
            ('database', STRING_TYPE),
            ('port', INT_TYPE),
            ('server', STRING_TYPE),
            ('ssl', BOOL_TYPE),
            ('verify', BOOL_TYPE)
        ]}
    ENTRY_FIELDS = [
        ('fields', ARRAY_TYPE),
        ('tags', HASH_TYPE)
    ]

    def __init__(self, path, root):
        """
        Constructor for the config object. The path value should represent an
        existing config file that the parser should read in. The root corresponds
        to the application defined root element indicating which config elements
        should be parsed form the config file.

        :param path: Path o the config file.
        :param root: Root element in the global section indicating all entries
                     to parse.
        """
        self.path = path
        self.root = root
        self.config = {}
        self.database = None

    def GetDatabase(self):
        """
        Retrieve the database configuration from the underlying config.

        :return: A tuple consisting of the database type and the respective configuration.
        """
        if not self.IsLoaded():
            self.Load()
        return self.database, self.config[self.database]

    def GetField(self, field):
        """
        Query the main fields list and return the type hint for the given field if it
        exists. The function will throw a KeyError in the event the field does not
        exist or is an invalid submission.

        :param field: String field to lookup in the config.
        :return: Type hint if the field is known.
        """
        if field is None:
            raise KeyError('Unknown field')
        if not self.IsLoaded():
            self.Load()
        return self.config['fields'][field]

    def GetRoot(self):
        """
        Returns the application root of the configuration file.

        :return: Application data from the root element.
        """
        if not self.IsLoaded():
            self.Load()
        if self.path is None:
            return {}
        return self.config[self.root]

    def GetTags(self, entity):
        """
        Lookup the tags for an entity if it is known. I fthe entity is not known a
        key error is returned. If the entity has no tags an empty list will be
        returned.

        :param entity: An entity to lookup in the config.
        :return: Tag list if the entity is known or an empty list if no tags are given.
        """
        if entity is None:
            raise KeyError('Unknown entity')
        if not self.IsLoaded():
            self.Load()
        if self.path is None:
            return {}
        return self.config[self.root][entity].get('tags', [])

    def IsLoaded(self):
        """
        Predicate checking if the cofnig has already been loaded.

        :return: True or False if the config has been loaded.
        """
        if self.path is None:
            return True
        return len(self.config) > 0

    def Load(self):
        """
        Load the configuration for the given file.

        On an error, this function will raise an error indicating what the failure
        was when it occurred.

        :return: None
        """
        if self.IsLoaded():
            return

        if not os.path.isfile(self.path):
            raise ConfigError('Config file does not exist')

        parser = ConfigParser()
        try:
            parser.read(self.path)
        except IOError:
            raise ConfigError('Failed to parse configuration')

        if not parser.has_section(self.GLOBAL_SECTION):
            raise InvalidConfigError('No global section')

        self.RequiredFields(parser, self.GLOBAL_SECTION, ['database', self.root])
        self.database = parser.get(self.GLOBAL_SECTION, 'database')

        if not self.database or self.database not in self.SUPPORTED_DATABASES:
            raise InvalidConfigError('Invalid or unsupported database value')

        self.config[self.database] = {}
        for field, hint in self.DATABASE_FIELDS[self.database]:
            if parser.has_option(self.database, field):
                try:
                    self.config[self.database][field] = ConvertValue(
                        parser.get(self.database, field),
                        hint=hint)
                except ConversionFailure:
                    raise InvalidConfigError("Invalid field '{}' expected type '{}'"
                        .format(field, hint))

        root = parser.get(self.GLOBAL_SECTION, self.root)
        if not root:
            raise InvalidConfigError("Missing config for root '{}".format(self.root))

        self.config['fields'] = {}
        self.config[self.root] = {}
        for field in root.split():
            self.config[self.root][field] = {}

        if len(self.config[self.root]) == 0:
            raise InvalidConfigError("Root element '{}' has no entries".format(self.root))

        fields = {}
        for entry in self.config[self.root].keys():
            self.RequiredFields(parser, entry, [k for k, v in self.ENTRY_FIELDS])
            self.config[self.root][entry] = {}

            for field, hint in self.ENTRY_FIELDS:
                if parser.has_option(entry, field):
                    try:
                        self.config[self.root][entry][field] = ConvertValue(
                            parser.get(entry, field),
                            hint=hint)
                    except ConversionFailure:
                        raise InvalidConfigError("Invalid field '{}' expected type '{}'"
                            .format(field, hint))

            for option in parser.options(entry):
                if option in self.config[self.root][entry]:
                    continue
                try:
                    self.config[self.root][entry][option] = ConvertValue(
                        parser.get(entry, option))
                except ConversionFailure:
                    raise InvalidConfigError("Invalid field '{}' unable to determine type"
                        .format(option))

            for field in self.config[self.root][entry]['fields']:
                if field not in fields:
                    fields[field] = []

        for field in fields.keys():
            if not parser.has_section(field):
                raise InvalidConfigError("Unknown field '{}' in config")
            for option in parser.options(field):
                if field not in self.config['fields']:
                    self.config['fields'][option] = self.ParseOption(parser, field, option)
                    fields[field].append(option)
                else:
                    raise InvalidConfigError("Duplicate field definition for '{}' in section '{}'"
                        .format(option, field))

        for entry, values in self.config[self.root].items():
            fieldList = []
            for field in values['fields']:
                if field in fields:
                    fieldList.extend(fields[field])
                else:
                    raise InvalidConfigError("Unknown field '{}' in entry '{}'".format(field, entry))
            values['fields'] = fieldList

    @staticmethod
    def ParseOption(parser, section, option):
        """
        Parse a given option as a defined type. This attempts to match the value of an option
        to a known type which can be coerced into a native value.

        :param parser: ConfigParser instance
        :param section: Section containing fields
        :param option: Option with a value representing a known coerseable type.
        :return:
        """
        if not parser.has_option(section, option):
            raise ConfigError("Option '{}' does not exist in section '{}'"
                .format(option, section))
        value = parser.get(section, option)
        if value in [Config.BOOL_TYPE, Config.FLOAT_TYPE,
                     Config.INT_TYPE, Config.STRING_TYPE]:
            return value
        raise InvalidConfigError("Invalid type '{}' for option '{}'".format(value, option))

    @staticmethod
    def RequiredFields(parser, section, fields):
        """
        Check a section for required fields. This function will raise an error if all
        required fields are not present within the given section.

        :param parser: ConfigParser instance
        :param section: Section containing fields
        :param fields: Array of fields (strings) which should be validated for existence.
        :return: None
        """
        for field in fields:
            if not parser.has_option(section, field):
                raise InvalidConfigError("Section '{}' missing required field '{}'"
                    .format(section, field))

    def Reload(self):
        """
        Clean the config state and reload the config from disk.

        :return: None
        """
        self.config = {}
        self.database = None
        self.Load()



import fcntl
import grp
import os
import platform
import pwd
import select
import subprocess
from .exceptions import ExecutorError


class Callbacks(object):
    """
    Wrapper to allow for the Executor setup o register callbacks for secondary
    commands on the argument parser.
    """

    def __init__(self, command):
        """
        Constructor.

        :param command: Name of the primary run command that should be treated as a
                        protected value.
        """
        self.command = command
        self.callbacks = {}
        self.parsers = None

    def __call__(self, parsers, callback):
        """
        Callable which accepts the ArgumentParser from the Executor and passes this
        wrapper back to the caller to register their sub-commands.

        :param parsers: ArgumentParser sub-parser.
        :param callback: Uer given callback for registering subcommands.
        :return:
        """
        self.parsers = parsers
        callback(self)

    def Register(self, name, callback, **kwargs):
        """
        Register a callback and add a name to it.

        :param name: Name of the sub-command callback.
        :param callback: Callbable object.
        :return: Sub-Parser object
        """
        if not callable(callback):
            raise ExecutorError('Callback is not callable')
        if name == self.command:
            raise ExecutorError("Cannot register a secondary '{}' command".format(name))
        for [k, _] in self.callbacks.items():
            if name == k:
                raise ExecutorError("Command '{}' already registered".format(name))
        self.callbacks[name] = callback
        parser = self.parsers.add_parser(name, **kwargs)
        parser.add_argument('--config', help='Path to the config file')
        return parser


def CloseDescriptor(fd):
    """
    Close a given descriptor. If the object has its own 'close' method that
    will be used otherwise everything will be passed to 'os.close'.

    :param fd: File descriptor or 'closeable' object.
    :return: None
    """
    try:
        if hasattr(fd, 'close'):
            fd.close()
        else:
            os.close(fd)
    except (IOError, OSError):
        pass


def Command(command, stderr=True, cwd=None):
    """
    Execute an external command with the given working directory. The result will
    be the STDOUT data and the return code. If 'stderr' is set to true the STDERR
    output will be piped to STDOUT otherwise it will be ignored.

    :param command: Command parameters to execute.
    :param stderr: Boolean indicating whether STDERR should be piped to STDOUT.
    :param cwd: Current working directory.
    :return: Tuple of process exitcode and STDOUT data.
    """
    process = None
    output = []
    try:
        process = subprocess.Popen(command,
            cwd=cwd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if stderr else os.devnull)

        while True:
            if process.poll() is not None:
                break
            for line in iter(process.stdout.readline, b''):
                if len(line.strip()) > 0:
                    output.append(line.strip())
    except KeyboardInterrupt:
        pass
    except OSError:
        raise

    return process.poll(), output


def GetGroupId(group):
    """
    On a Linux system attempt to get the GID value for a give 'group'. This uses
    a system call to obtain this data.

    :param group: Group name to lookup.
    :return: GID value if the group is found, otherwise None
    """
    if isinstance(group, int):
        return group
    if platform.system() != 'Linux':
        raise RuntimeError('GID lookup not support outside of Linux')
    try:
        return grp.getgrnam(group).gr_gid
    except KeyError:
        return None


def GetUserId(user):
    """
    On a Linux system attempt to get the UID value for a give 'user'. This uses
    a system call to obtain this data.

    :param user: User name to lookup.
    :return: UID value if the user is found, otherwise None
    """
    if isinstance(user, int):
        return user
    if platform.system() != 'Linux':
        raise RuntimeError('UID lookup not support outside of Linux')
    try:
        return pwd.getpwnam(user).pw_uid
    except KeyError:
        return None


def RedirectStream(source, target=None):
    """
    Redirect a source file descriptor to a new target file descriptor. If no target
    is specified the source will be redirected to the /dev/null object.

    :param source: Source file descriptor
    :param target: Target file descriptor. If None is provided /dev/null is used.
    :return: None
    """
    if target is None:
        target = os.open(os.devnull, os.O_RDWR)
    else:
        target = target.fileno()
    os.dup2(target, source.fileno())


def Select(rds, wrts, timeout, logger=None):
    """
    Select on the given sockets for events at the given timeout period. This is a wrapper
    around the system call select for handling a set of file descriptors. The function will
    return None in the event of a failure of the resulting list of descriptors which an
    event occurred on.

    :param rds: Set of reader descriptors.
    :param wrts: Set of writer descriptors.
    :param timeout: Timeout value to wait for an event.
    :param logger: Optional logger instance in the event of errors.
    :return: None if a failure occurred, or the descriptors which an event occurred.
    """
    if not isinstance(rds, list):
        rds = [rds]
    if not isinstance(wrts, list):
        wrts = [wrts]

    try:
        res = select.select(rds, wrts, [], float(timeout))
    except select.error as e:
        if logger:
            logger.error('Select error: {}'.format(e))
        return None
    except os.error as e:
        if logger:
            logger.error('OSError: [{}] {}'.format(e.errno, os.strerror(e.errno)))
        return None
    except KeyboardInterrupt:
        return []

    return res


def SetNonBlocking(fd):
    """
    Set file descriptors to non-blocking.

    :param fd: File descriptor object
    :return: None
    """
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)


def SetProcessOwner(user, group, logger=None):
    """
    Set the given user and group as the process owner. The current process owner
    must have access to set the new user/group combination. This is usually only
    used in the event root needs to drop privileges to a non-privileged user after
    a fork has occurred.

    :param user: Integer representing the new user owner.
    :param group: Integer representing the new group owner.
    :param logger: Optional logger instance in the event of errors.
    :return: None
    """
    try:
        if user is not None:
            os.setuid(user)
    except OSError as e:
        if logger:
            logger.error("Failed to set process user '{}': [{}] {}".format(
                user, e.errno, os.strerror(e.errno)))
    try:
        if group is not None:
            os.setgid(group)
    except OSError as e:
        if logger:
            logger.error("Failed to set process group '{}': [{}] {}".format(
                group, e.errno, os.strerror(e.errno)))


def SetProcessUmask(umask, logger=None):
    """
    Set the process umask.

    :param umask: Selected umask.
    :param logger: Optional logger instance in the event of errors.
    :return: None
    """
    try:
        os.umask(umask)
    except OSError as e:
        if logger:
            logger.error('Failed to set umask: [{}] {}'.format(
                e.errno, os.strerror(e.errno)))



try:
    import influxdb
    INFLUXDB_SUPPORTED = True
except ImportError:
    INFLUXDB_SUPPORTED = False


class Database(object):

    def Close(self):
        raise NotImplementedError

    def Initialize(self):
        raise NotImplementedError

    def Write(self, metrics):
        raise NotImplementedError


class InfluxDatabase(Database):

    def __init__(self, config, precision='ms', logger=None):
        """
        InfluxDB database constructor.

        This class will throw a RuntimeError if InfluxDB is not supported in the
        current environment.

        :param config: Database configuration from the main configuration object.
        :param precision: Timestamp precision. Defaults to ms.
        :param logger: Logger instance to use otherwise logging will be ignored.
        """
        if not INFLUXDB_SUPPORTED:
            raise RuntimeError('InfluxDB is not installed')

        self.config = config
        self.precision = precision
        self.logger = logger
        self.handle = None

    def Close(self):
        """
        Close the InfluxDB connection if one is open.

        This function will only throw in the event that InfluxDB is not supported
        otherwise it is safe to call during shutdown.

        :return: None
        """
        if not INFLUXDB_SUPPORTED:
            raise RuntimeError('InfluxDB is not installed')
        if self.handle:
            try:
                self.handle.close()
            except Exception as e:
                if self.logger:
                    self.logger.warning('Failed to close influx db connection: {}'.format(e))

    def Initialize(self):
        """
        Initialize the InfluxDB connection.

        This class will throw a RuntimeError if InfluxDB is not supported in the
        current environment.

        :return: True on success or False if the connection fails.
        """
        if not INFLUXDB_SUPPORTED:
            raise RuntimeError('InfluxDB is not installed')
        try:
            self.handle = influxdb.InfluxDBClient(
                host=self.config['server'],
                port=self.config['port'],
                ssl=self.config['ssl'],
                verify_ssl=self.config['verify'],
                database=self.config['database'])
        except KeyError as e:
            # If any of the configuration options are missing we need to trigger a fatal
            # error because we cannot recover from this.
            raise RuntimeError("Missing configuration option '{}'".format(e.args[0]))
        except Exception as e:
            if self.logger:
                self.logger.error('Failed to initiate influx db connection: {}'.format(e))
            self.handle = None
            return False

        return True

    def Write(self, metrics):
        """
        Write metrics to the InfluxDB database.

        This class will throw a RuntimeError if InfluxDB is not supported in the
        current environment.

        :param metrics: metrics object generated that should be sent to InfluxDB.
        :return: True on success or False on failure.
        """
        if not INFLUXDB_SUPPORTED:
            raise RuntimeError('InfluxDB is not installed')
        if not self.handle and not self.Initialize():
            return False
        self.handle.write_points(metrics, time_precision=self.precision)
        return True



import os
import sys
from .utils import GetGroupId, GetUserId, RedirectStream, SetProcessOwner, SetProcessUmask


class Daemonize(object):

    DEFAULT_UMASK = 0
    EXIT_FAILURE = 1
    EXIT_SUCCESS = 0

    def __init__(self, launch, logger, user=None, group=None,
                 stdin=None, stdout=None, stderr=None,
                 umask=DEFAULT_UMASK):
        """
        Constructor for the daemon context.

        :param launch: Boolean indicating if the process should fork.
        :param logger: Logger instance shared from the caller.
        :param user: Optional user which should own the forked process. Only used
                     if the process daemonizes.
        :param group: Optional group which should own the forked process. Only used
                     if the process daemonizes.
        :param stdin: New STDIN target file descriptor.
        :param stdout: New STDOUT target file descriptor.
        :param stderr: New STDERR target file descriptor.
        :param umask: Process umask.
        """
        self.launch = bool(launch)
        self.logger = logger
        self.umask = umask
        self.gid = GetGroupId(group) if group else None
        self.uid = GetUserId(user) if user else None
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.launched = False

    def __enter__(self):
        """
        Initialize the daemon process. If the process calls for daemonizing this
        context will result in the application forking from the parent. It performs
        a standard double-fork and will call the relevant setUID/GUID/etc functions
        determined by the constructor.

        :return: None
        """
        if self.logger:
            if self.uid:
                self.logger.debug('Launching with UID={}', self.uid)
            if self.gid:
                self.logger.debug('Launching with GID={}', self.gid)
            self.logger.debug('--- START DAEMON CONTEXT ---')

        self.pid = os.getpid()
        if self.launch:
            self.__Daemonize()

        SetProcessUmask(self.umask, logger=self.logger)
        SetProcessOwner(self.uid, self.gid, logger=self.logger)

    def __exit__(self, exc, value, tb):
        """
        Shuts down the process. If an exception resulted in the shutdown the resulting
        exit code will indicate a failure otherwise an exit code of '0' is returned
        indicating success.

        :param exc: Exception type
        :param value: Exception message
        :param tb: Traceback of the exception
        :return: None
        """
        self.logger.debug('--- STOP DAEMON CONTEXT ---')
        if exc:
            sys.exit(self.EXIT_FAILURE)
        else:
            sys.exit(self.EXIT_SUCCESS)

    def __Daemonize(self):
        """
        Daemonize the application by doing a standard double fork. This will result
        in the UID/GID/etc options being set after the fork.

        :return: None
        """
        if self.launch and not self.launched:
            # Don't fork if our parent PID is 1. This usually indicates that we
            # are already running as a Daemon or at least launched by init.
            if os.getppid() == 1:
                return

            try:
                self.__Fork('first fork')
            except OSError:
                exit(self.EXIT_FAILURE)

            os.setsid()

            try:
                self.__Fork('second fork')
            except OSError:
                exit(self.EXIT_FAILURE)

            self.launched = True
            RedirectStream(sys.stdin, self.stdin)
            RedirectStream(sys.stdout, self.stdout)
            RedirectStream(sys.stderr, self.stderr)

    def __Fork(self, error=None):
        """
        Wrapper around the form call to handle exiting from the parent cleanly.

        :param error: Error message indicating which fork this call represents this will
                      get added to an optional log message in the event of a failure.
        :return: None
        """
        try:
            if os.fork() > 0:
                exit(self.EXIT_SUCCESS)
        except OSError as e:
            if error and self.logger:
                self.logger.error("Fork error for '{}': [{}] {}".format(
                    error, e.errno, os.strerror(e.errno)))
            elif self.logger:
                self.logger.error('Failed to fork parent: [{}] {}'.format(
                    e.errno, os.strerror(e.errno)))
            raise



import setuptools

try:
    with open('README.md', 'rb') as handle:
        description = handle.read()
except IOError as e:
    print('Failed to read README file: {}'.format(e))
    exit(1)

# Packaging tutorial: https://packaging.python.org/tutorials/packaging-projects/
# Package classifiers: https://pypi.org/classifiers/

setuptools.setup(
    name='PyMonitorLib',
    version='0.3.0',
    author='Daniel Weiner',
    author_email='info@phantomnet.net',
    description='Library for creating simple interval processes. This is especially '
                'useful for interval based monitoring application that generate '
                'telemetry at a set interval.',
    long_description=description.decode('utf-8'),
    long_description_content_type='text/markdown',
    url='https://github.com/Aprelius/PyMonitorLib',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



from .config import Config, ConfigError, ConversionFailure, ConvertBoolean, ConvertHashType, \
    ConvertValue, InvalidConfigError
from .daemon import Daemonize
from .database import Database, InfluxDatabase
from .exceptions import MessageError
from .executor import Executor, Execute
from .metrics import Metric, MetricPipeline
from .result import Result
from .utils import CloseDescriptor, Command, GetGroupId, GetUserId, Select, SetNonBlocking



class MessageError(Exception):
    message = 'Whoops! Something went wrong'

    def __init__(self, message=None, *args, **kwargs):
        self.message = message or self.message
        if args:
            self.message = message.format(*args)
        elif kwargs:
            self.message = message.format(**kwargs)


class ExecutorError(MessageError):
    message = 'An error occurred in the Executor'
